@@ -1,3 +1,15 @@
+/* SLIRP preprocessing extension {{{
+ * This code augments the S-Lang preprocessor, giving coders the capability 
+ * to define their own preprocessing tokens in either C or S-Lang code; it
+ * forms the basis of the SLIRP annotation mechanism, but could also be
+ * used separately.  In S-Lang scope new preprocessor tokens are registered
+ * via preproc_handler_add(): this associates the token with a callback
+ * (S-Lang function), indicates how blocks opened with the token must be
+ * terminated (e.g. is it a single-line token, or does it require "#end"),
+ * and optionally registers an arbitrary number of arguments to pass to
+ * the callback each time it's invoked.
+ * }}} */ 
+
 #include <stdio.h>
 #include <slang.h>
 #include <string.h>
@@ -6,7 +18,6 @@
 #include "preproc-module.h"
 
 #if SLANG_VERSION < 20000
-#define SLprep_set_flags(pp,f)	pp->flags != f
 #define SLprep_Type		SLPreprocess_Type
 #define SLprep_delete(pp)	SLfree((char*)pp)
 extern int			(*_SLprep_eval_hook) (char *);  /* _slang.h */
@@ -14,7 +25,7 @@ extern int			(*_SLprep_eval_hook) (char *);  /* _slang.h */
 
 /* Local types and variables {{{ */
 
-#define MAX_FILE_LINE_LEN 256
+#define MAX_FILE_LINE_LEN 256	/* matches sltoken.c */
 
 typedef struct _PassThru PassThru;
 struct _PassThru {
@@ -29,6 +40,8 @@ struct _Handler_Entry {
    size_t open_pat_len;		/* optimization: avoids unnecc strlen calls */
    char *close_pat;		/* close block token */
    size_t close_pat_len;
+#define ALLOW_UNTERM	0x01
+   unsigned int flags;
    SLprep_Handler handler;
    PassThru *pt_head;		/* stuff which may pass thru sans complaint */
    PassThru *pt_tail;
@@ -165,9 +178,19 @@ static char *read_from_file (SLang_Load_Type *x) /* {{{ */
 
 	if (fgets (c->buf, MAX_FILE_LINE_LEN, c->fp) == NULL) {
 
-	   if (Current != NULL)
-		SLang_verror (SL_SYNTAX_ERROR, "Unterminated <#%s> directive",
-							Current->open_pat);
+	   /* Have we hit EOF without closing current prepr block? */
+	   if (Current != NULL) {
+
+			/* ... ok if block can be a single line */
+		if (Current->flags & ALLOW_UNTERM) {
+		   Current->flags ^=  ALLOW_UNTERM;
+		   Current = NULL;
+		}
+		else	/* ... otherwise not ok */
+		   SLang_verror (SL_SYNTAX_ERROR,
+			"Unterminated <#%s> directive", Current->open_pat);
+	   }
+
 	   return NULL;
 	}
 
@@ -455,8 +478,8 @@ static int slang_args_extract(unsigned int nargs, SLang_Any_Type ***pargs)
 
 static void slang_args_free(unsigned int nargs, SLang_Any_Type **args)
 {
-   while (nargs > 0)
-	SLang_free_anytype(args[--nargs]);
+   while (nargs--)
+	SLang_free_anytype(args[nargs]);
    SLfree((char*)args);
 }
 
@@ -665,6 +688,12 @@ static void deactivate()
    SLns_Load_File_Hook = Previous_Load_File_Ns_Hook;
 }
 
+static void allow_unterminated()
+{
+   if (Current != NULL)			  /* support single-line blocks for  */
+	Current->flags |= ALLOW_UNTERM;	  /* tokens otherwise requiring #end */
+}
+
 #define I SLANG_INT_TYPE
 #define V SLANG_VOID_TYPE
 #define S SLANG_STRING_TYPE
@@ -674,6 +703,7 @@ static SLang_Intrin_Fun_Type Intrin_Funcs [] =
    MAKE_INTRINSIC_0("preproc_activate", activate, V),
    MAKE_INTRINSIC_0("preproc_deactivate", deactivate, V),
    MAKE_INTRINSIC_0("preproc_handler_add", handler_add_wrapper, V),
+   MAKE_INTRINSIC_0("preproc_allow_unterminated", allow_unterminated, V),
    MAKE_INTRINSIC_0("preproc_handler_remove", handler_remove_wrapper, V),
    MAKE_INTRINSIC_0("preproc_handler_exists", handler_exists_wrapper, V),
    MAKE_INTRINSIC_0("preproc_handler_passthru", handler_passthru_wrapper, V),
@@ -693,7 +723,7 @@ static SLang_Intrin_Var_Type Intrin_Vars[] =
 /*}}}*/
 
 /* Module interface {{{ */
-static int preproc_callback_hook (char *buf, char prep_char)
+static int preproc_callback_hook (char *buf, char prep_char) /*{{{*/
 {
    int ret = 1;			/* let line pass through, sans inspection */
 
@@ -713,6 +743,19 @@ static int preproc_callback_hook (char *buf, char prep_char)
 		else
 		   return 1;		/* malformed preprocessor block */
 	   }
+	   else if (Current->flags & ALLOW_UNTERM) {
+
+		/* Terminate token block implicitly ... */
+		if (handler_invoke(NULL, e) != 0)
+		   return 1;
+
+		/* turn off permission for next block 2b unterminated ... */
+		Current->flags ^= ALLOW_UNTERM;
+		Current = NULL;
+
+		/* ... and see if this line needs preprocessing */
+		return preproc_callback_hook(--buf, prep_char);
+	   }
 	   else if (passthru(e, buf))
 		buf--;
 	   else
@@ -738,11 +781,11 @@ static int preproc_callback_hook (char *buf, char prep_char)
 
    return ret;
 
-}
+} /*}}}*/
 
 SLANG_MODULE(preproc);
 
-int init_preproc_module_ns (char *ns_name)
+int init_preproc_module_ns (char *ns_name) /*{{{*/
 {
    SLang_NameSpace_Type *ns;
    
@@ -757,7 +800,7 @@ int init_preproc_module_ns (char *ns_name)
 	return -1;
 
    return 0;
-}
+} /*}}}*/
 
 void deinit_preproc_module (void)
 {
