@@ -1,4 +1,4 @@
-static void Slirp_usage(int i, int last, int vectorized) /* {{{ */
+static void Slirp_usage(int i, int last, int flags) /* {{{ */
 {
    char *indent;
    int npop = SLstack_depth();
@@ -13,6 +13,9 @@ static void Slirp_usage(int i, int last, int vectorized) /* {{{ */
    do
 	SLang_verror(SLEU, (char*)"%s%s", indent, usage_strings[i++]);
    while (i < last);
-   if (vectorized)
+   if (flags & 0x2)
+	SLang_verror(SLEU,
+		(char*)"\tThis function has been vectorized and parallelized.");
+   else if (flags & 0x1)
 	SLang_verror(SLEU, (char*)"\tThis function has been vectorized.");
 } /* }}} */
